@@ -1,6 +1,6 @@
 /* Second stage boot loader
    
-   Copyright (C) 1996 Pete A. Zaitcev
+   Coyright (C) 1996 Pete A. Zaitcev
    		 1996 Maurizio Plaza
    		 1996 David S. Miller
    		 1996 Miguel de Icaza
@@ -19,22 +19,42 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+   USA.  */
 
 /* TODO: This file is a good candidate for rewrite from scratch.  */
 
 #include <silo.h>
-#include <asm/page.h>
-#include <linux/elf.h>
+#include <elf.h>
 #include <stringops.h>
 
 #ifndef NULL
 #define NULL (void *)0
 #endif
 
+
+struct HdrS_struct {
+	char magic[4];
+	unsigned int linux_ver;
+	unsigned short ver;
+	unsigned short root_flags;
+	unsigned short root_dev;
+	unsigned short ram_flags;
+	unsigned int ramdisk_image;
+	unsigned int ramdisk_size;
+	/* 0x201 */
+	unsigned int reboot_cmd_ptr_high;
+	unsigned int reboot_cmd_ptr_low;
+	/* 0x202 */
+	unsigned int bootstr_info_ptr_high;
+	unsigned int bootstr_info_ptr_low;
+	/* 0x301 */
+	unsigned long long ramdisk_image64;
+};
+
 /* This has to be first initialized variable in main.c */
 
-extern unsigned char silo_conf[256];
+extern char silo_conf[256];
 extern unsigned char silo_conf_part, silo_conf_parts[32], raid_dsk_number;
 
 int useconf = 0;
@@ -63,13 +83,13 @@ static char other_device [512];
 static int reboot = 0;
 static int floppyswap = 0;
 int tab_ambiguous = 0;
-unsigned int linux_end = 0;
 int other_part = -1;
 int solaris = 0;
 int other = 0;
 char *password = 0;
+int initrd_can_do_64bit_phys = 0;
 
-void parse_name (char *, int, char **, int *, char **);
+static void parse_name (char *, int, char **, int *, char **);
 
 static char *next_tok (char *t) {
     while (*t++); /* That was easy */
@@ -92,7 +112,7 @@ static int tab_complete(void) {
     if (p && *p >= '1' && *p <= '8' && !p[1])
 	defpart = *p - '0';
     else {
-	fatal("\nDefault partition could not be found");
+	silo_fatal("\nDefault partition could not be found");
 	free(r);
 	return 1;
     }
@@ -146,8 +166,9 @@ static int tab_complete(void) {
     } else {
 	if (!device) device = cfg_get_strg (0, "device");
 
-	if (load_file(device, part, kname, (unsigned char *) 0x4000,
-		(unsigned char *) &_start, &image_len, LOADFILE_LS_MATCH|LOADFILE_QUIET, 0))
+	if (silo_load_file(device, part, kname, (unsigned char *) 0x4000,
+			   (unsigned char *) &_start, &image_len,
+			   LOADFILE_LS_MATCH|LOADFILE_QUIET, 0))
 	    if (do_ls((unsigned char *)0x4000, &tab_ambiguous))
 		ret = 1;
     }
@@ -172,7 +193,8 @@ static void maintabfunc (void)
     return;
 }
 
-void parse_name (char *imagename, int defpart, char **device, int *part, char **kname)
+static void parse_name (char *imagename, int defpart, char **device,
+			int *part, char **kname)
 {
     static char parsebuf[1024];
 
@@ -182,7 +204,7 @@ void parse_name (char *imagename, int defpart, char **device, int *part, char **
     *kname = 0;
     if (prom_vers == PROM_V0) {
         static char v0_buffer[20];
-        *kname = v0_device (imagename);
+        *kname = silo_v0_device(imagename);
         if (*kname) {
             memcpy (v0_buffer, imagename, *kname - imagename + 1);
             v0_buffer [*kname - imagename + 1] = 0;
@@ -230,7 +252,7 @@ void parse_name (char *imagename, int defpart, char **device, int *part, char **
 	*kname = 0;
 }
 
-void check_initrd (char *label, int defpart, char *defdevice)
+static void check_initrd (char *label, int defpart, char *defdevice)
 {
     char *p;
     
@@ -245,7 +267,7 @@ void check_initrd (char *label, int defpart, char *defdevice)
     }
 }
 
-int dig_into_params (char *params)
+static int dig_into_params (char *params)
 {
     char *p, *q;
     char *last_root = 0;
@@ -299,14 +321,14 @@ int dig_into_params (char *params)
     return root_found;
 }
 
-void check_password(char *str)
+static void check_password(char *str)
 {
     int i;
     
     for (i = 0; i < 3; i++) {
 	printf ("\n%sassword: ", str);
 	passwdbuff[0] = 0;
-	cmdedit ((void (*)(void)) 0, 1);
+	silo_cmdedit((void (*)(void)) 0, 1);
 	printf ("\n");
 	if (!strcmp (password, passwdbuff))
 	    return;
@@ -317,7 +339,7 @@ void check_password(char *str)
     prom_halt();
 }
 
-char *ls_options(char *parms)
+static char *ls_options(char *parms)
 {
     char *p = parms;
     ls_opt = 0;
@@ -337,7 +359,8 @@ char *ls_options(char *parms)
     return p;
 }
 
-int get_params (char **device, int *part, char **kname, char **proll, char **params)
+static int get_params (char **device, int *part, char **kname, char **proll,
+		       char **params)
 {
     int defpart = -1;
     char *defdevice = 0;
@@ -368,7 +391,7 @@ int get_params (char **device, int *part, char **kname, char **proll, char **par
     pause_after = 0;
     reboot = 0;
     *proll = 0;
-    cmdinit ();
+    silo_cmdinit();
     *params = "";
     if (useconf) {
 	defdevice = cfg_get_strg (0, "device");
@@ -385,7 +408,7 @@ int get_params (char **device, int *part, char **kname, char **proll, char **par
 	    if (p && *p)
 	        timeout = atoi (p);
 	    if (no_prom_args) p = 0;
-	    else p = get_bootargs (0);
+	    else p = silo_get_bootargs(0);
 	    if (p) while (*p == ' ') p++;
 	    if (p && *p) {
 	        for (q = p; *q && *q != ' '; q++);
@@ -460,7 +483,7 @@ int get_params (char **device, int *part, char **kname, char **proll, char **par
 	if (!imagename) {
 	    if ((!*cbuff || (timeout > 0 && timer_status < 0)) && !tabbedout)
                 printf ("boot: ");
-	    cmdedit (maintabfunc, 0);
+	    silo_cmdedit(maintabfunc, 0);
 	    if (*cbuff == ' ') {
 		for (p = cbuff; *p == ' '; p++);
 		q = cbuff;
@@ -585,7 +608,7 @@ int get_params (char **device, int *part, char **kname, char **proll, char **par
     		    if (defdevice)
     		        oth_device = defdevice;
     		    else
-    		        oth_device = bootdevice;
+    		        oth_device = silo_disk_get_bootdevice();
     		}
     		strcpy (other_device, oth_device);
 	    	p = cfg_get_strg (label, "bootblock");
@@ -611,7 +634,7 @@ int get_params (char **device, int *part, char **kname, char **proll, char **par
         if (first) {
             first = 0;
 	    if (no_prom_args) p = 0;
-	    else p = get_bootargs (0);
+	    else p = silo_get_bootargs(0);
 	    if (p) while (*p == ' ') p++;
 	    if (p && *p) {
 	        for (q = p; *q && *q != ' '; q++);
@@ -627,7 +650,7 @@ int get_params (char **device, int *part, char **kname, char **proll, char **par
 	}
 	if (!imagename) {
             printf ("boot: ");
-	    cmdedit ((void (*)(void)) 0, 0);
+	    silo_cmdedit((void (*)(void)) 0, 0);
 	    if (*cbuff == ' ') {
 		for (p = cbuff; *p == ' '; p++);
 		q = cbuff;
@@ -732,7 +755,7 @@ int get_params (char **device, int *part, char **kname, char **proll, char **par
     	if (*part != -2) {
 	    other_part = *part;
 	    if (!*device)
-	    	strcpy (other_device, bootdevice);
+	    	strcpy (other_device, silo_disk_get_bootdevice());
 	    else
 	    	strcpy (other_device, *device);
 	    p = strstr (*params, "bootblock=");
@@ -830,21 +853,25 @@ int get_params (char **device, int *part, char **kname, char **proll, char **par
     return 0;
 }
 
-void initrd_lenfunc (int len, char **filebuffer, char **filelimit)
+static void initrd_lenfunc (int len, char **filebuffer, char **filelimit)
 {
+    extern unsigned long long initrd_phys;
+
     initrd_start = memory_find ((len + 16383) & ~16383);
     if (!initrd_start) {
-        fatal ("You do not have enough continuous available memory for such initial ramdisk.");
+        silo_fatal("You do not have enough continuous available memory "
+		   "for such initial ramdisk.");
         prom_halt ();
     }
     initrd_size = len;
-    printf ("Loading initial ramdisk....\n");
     *filebuffer = initrd_start;
     *filelimit = initrd_start + ((len + 16383) & ~16383);
+    printf("Loading initial ramdisk (%d bytes at 0x%Lx phys, 0x%x virt)...\n", len,
+	   initrd_phys, initrd_start);
 }
 
-int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
-    unsigned *pstart, char *image_name)
+static int parse_executable (unsigned char *base, int image_len, unsigned int *poff,
+			     int *plen, unsigned *pstart, char *image_name)
 {
     int isfile = 0;
     union {
@@ -861,7 +888,7 @@ int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
     /* the kernel) so we check if it is an executable file, either */
     /* an a.out or an elf binary */
 
-    hp.b = base;
+    hp.b = (char *)base;
     if (hp.a->magic == 0x01030107) {
     	if (solaris) {
     	    printf ("\nYour Solaris `ufsboot' is not an ELF image. Try again.\n");
@@ -879,7 +906,7 @@ int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
 		hp.e->e_ident[EI_MAG3] == ELFMAG3) {
 
         if (hp.e->e_ident[EI_DATA] != ELFDATA2MSB) {
-	    fatal ("Image is not a MSB ELF");
+	    silo_fatal("Image is not a MSB ELF");
 	    prom_halt ();
 	}
 	if (hp.e->e_ident[EI_CLASS] == ELFCLASS32) {
@@ -887,7 +914,8 @@ int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
 
 	    p = (Elf32_Phdr *) (hp.b + hp.e->e_phoff);
 	    if (p->p_type != PT_LOAD) {
-		fatal ("Cannot find a loadable segment in your ELF image");
+		silo_fatal("Cannot find a loadable segment in your "
+			   "ELF image");
 		prom_halt ();
 	    }
 	    if (solaris) {
@@ -897,7 +925,8 @@ int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
 	        for (i = 0; i < hp.e->e_phnum; i++, p++) {
 	            if (p->p_vaddr < 0x4000 + image_len ||
 				p->p_vaddr + p->p_memsz >= sa) {
-	        	fatal("Unable to handle your Solaris `ufsboot' bootloader.");
+	        	silo_fatal("Unable to handle your "
+				   "Solaris `ufsboot' bootloader.");
 	            	prom_halt ();
 	            }
 	            memcpy ((char *)p->p_vaddr,
@@ -925,7 +954,8 @@ int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
 				p->p_filesz = n + q->p_filesz;
 				p->p_memsz = n + q->p_memsz;
 	                } else {
-	                    fatal("Multiple loadable segments in your ELF image");
+	                    silo_fatal("Multiple loadable segments in "
+				       "your ELF image");
 	                    prom_halt();
 	                }
 	            }
@@ -942,7 +972,8 @@ int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
 
 	    p = (Elf64_Phdr *) (hp.b + hp.f->e_phoff);
 	    if (p->p_type != PT_LOAD) {
-		fatal ("Cannot find a loadable segment in your ELF image");
+		silo_fatal("Cannot find a loadable segment in your "
+			   "ELF image");
 		prom_halt ();
 	    }
 	    if (solaris) {
@@ -953,7 +984,8 @@ int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
 		    if (p->p_vaddr < 0x4000 + image_len ||
 			 p->p_vaddr + p->p_memsz >= sa) {
 
-			 fatal("Unable to handle your Solaris `ufsboot' bootloader.");
+			 silo_fatal("Unable to handle your "
+				    "Solaris `ufsboot' bootloader.");
 			 prom_halt ();
  		    }
 		    memcpy ((Elf64_Addr *)(long)(p->p_vaddr),
@@ -979,7 +1011,8 @@ int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
 			    p->p_filesz = n + q->p_filesz;
 			    p->p_memsz = n + q->p_memsz;
 			} else {
-			    fatal("Multiple loadable segments in your ELF image");
+			    silo_fatal("Multiple loadable segments "
+				       "in your ELF image");
 			    prom_halt();
 			}
 		    }
@@ -1009,25 +1042,22 @@ int bootmain (void)
     int len = 0, image_len;
     char *kname, *params, *device;
     char *proll = 0;
-    char *image_base = (char *) 0x4000;
+    unsigned char *image_base = (unsigned char *) 0x4000;
     char *kernel_params;
     int part;
     int isfile, fileok = 0;
-    char *p;
     unsigned int ret_offset = 0;
     char *params_device = 0;
     int silo_conf_partition;
 
     prom_ranges_init ();
     get_idprom();
-    architecture = get_architecture ();
-    strcpy (given_bootargs, get_bootargs (1));
-    strcpy (my_bootargs, get_bootargs (0));
-
-    printf (" ");
+    architecture = silo_get_architecture();
+    strcpy (given_bootargs, silo_get_bootargs(1));
+    strcpy (my_bootargs, silo_get_bootargs(0));
 
 #ifndef TFTP
-    if (diskinit () == -1)
+    if (silo_diskinit() == -1)
 	prom_halt ();
 #endif
 	
@@ -1057,11 +1087,14 @@ int bootmain (void)
     if (*silo_conf && silo_conf_partition >= 1 && silo_conf_partition <= 8) {
 	int len;
 	solaris = 0;
-	fileok = load_file (0, silo_conf_partition, silo_conf, (unsigned char *) 0x4000, (unsigned char *) &_start, &len, LOADFILE_GZIP, 0);
+	fileok = silo_load_file(0, silo_conf_partition, silo_conf,
+				(unsigned char *) 0x4000,
+				(unsigned char *) &_start,
+				&len, LOADFILE_GZIP | LOADFILE_NO_ROTATE, 0);
 	if (!fileok || (unsigned) len >= 65535)
 	    printf ("\nCouldn't load %s\n", silo_conf);
 	else {
-	    if (!cfg_parse (silo_conf, (unsigned char *) 0x4000, len)) {
+	    if (!cfg_parse (silo_conf, (char *) 0x4000, len)) {
 		char *p, *q;
 		int len = 0;
 		int defpart = -1;
@@ -1079,7 +1112,10 @@ int bootmain (void)
 			if (!device)
 			    device = cfg_get_strg (0, "device");
 			solaris = 0;
-			if (load_file (device, part, kname, (unsigned char *) 0x4000, (unsigned char *) &_start, &len, LOADFILE_GZIP, 0)) {
+			if (silo_load_file(device, part, kname,
+					   (unsigned char *) 0x4000,
+					   (unsigned char *) &_start, &len,
+					   LOADFILE_GZIP, 0)) {
 			    *(unsigned char *) (0x4000 + len) = 0;
 			    printf ("\n");
 			    print_message ((char *) 0x4000);
@@ -1105,7 +1141,7 @@ int bootmain (void)
 	"Type [prompath;]part/path_to_image [parameters] on the prompt\n"
 	  "E.g. /iommu/sbus/espdma/esp/sd@3,0;4/vmlinux root=/dev/sda4\n"
 		"or 2/vmlinux.live (to load vmlinux.live from 2nd partition of boot disk)\n");
-
+try_again:
     isfile = 0;			/* RC = 0 invalid file or not an executable */
     while (!isfile) {
 	switch (get_params (&device, &part, &kname, &proll, &params)) {
@@ -1122,7 +1158,7 @@ int bootmain (void)
 	    break;
 
 	if (solaris) {
-	    char *p = seed_part_into_device ((!device || !*device) ? bootdevice : device, part);
+	    char *p = seed_part_into_device ((!device || !*device) ? silo_disk_get_bootdevice() : device, part);
 	    strcpy (sol_params, p);
 	    params_device = strchr (sol_params, 0) + 1;
 	    strcpy (params_device, kname);
@@ -1137,18 +1173,21 @@ int bootmain (void)
 		printf ("\nNeither \"other\" nor \"solaris\" are compatible with proll\n");
 		continue;
 	    }
-	    if (!load_file (device, part, proll, (unsigned char *) 0x4000,
-			(unsigned char *) 0x40000, &image_len, LOADFILE_GZIP, 0)) {
+	    if (!silo_load_file(device, part, proll, (unsigned char *) 0x4000,
+				(unsigned char *) 0x40000, &image_len,
+				LOADFILE_GZIP, 0)) {
 	        printf ("\nProll not found.... try again\n");
 		continue;
 	    }
-            if (!parse_executable ((char *)0x4000, image_len, &off, &len, NULL, proll))
+            if (!parse_executable ((unsigned char *)0x4000, image_len, &off,
+				    &len, NULL, proll))
 		continue;
             memcpy ((char *) 0x4000, ((char *) 0x4000) + off, len);
 
-	    image_base = (char *) 0x40000;
-	    if (!load_file (device, part, kname, image_base,
-			(unsigned char *) &_start, &image_len, LOADFILE_GZIP, 0)) {
+	    image_base = (unsigned char *) 0x40000;
+	    if (!silo_load_file(device, part, kname, image_base,
+				(unsigned char *) &_start, &image_len,
+				LOADFILE_GZIP, 0)) {
 	        printf ("\nImage not found.... try again\n");
 		continue;
 	    }
@@ -1158,11 +1197,38 @@ int bootmain (void)
 	    ret_offset = 0x4000;
 
 	} else {
-	    image_base = (unsigned char *) 0x4000;
-	    if (!load_file (device, part, kname, image_base,
-			(unsigned char *) &_start, &image_len,
-			load_cmd == CMD_LS ? LOADFILE_LS : LOADFILE_GZIP, 0)) {
+	    unsigned char *image_end = (unsigned char *)&_start;
+
+	    /* See if we can use some extra memory for the kernel */
+	    if (!load_cmd) {
+		unsigned int size;
+		unsigned char *mem;
+
+		size = 0x800000;
+		mem = (unsigned char *)image_memory_find(size);
+
+		if (!mem) {
+		    size = 0x400000;
+		    mem = (unsigned char *)image_memory_find(size);
+		}
+
+		if (mem) {
+		    image_base = mem;
+		    image_end = image_base + size - 0x4000;
+		    printf("Allocated %d Megs of memory at 0x%x for kernel\n",
+			   size >> 20, image_base - 0x4000);
+		}
+	    }
+
+	    if (!silo_load_file(device, part, kname, image_base, image_end,
+				&image_len, ((load_cmd == CMD_LS) ?
+					     LOADFILE_LS :
+					     LOADFILE_GZIP), 0)) {
 		printf ("\nImage not found.... try again\n");
+
+		if (!load_cmd)
+		    image_memory_release();
+
         	continue;
             }
 
@@ -1177,8 +1243,8 @@ int bootmain (void)
         	continue;
 	    }
 
-            isfile = parse_executable (image_base, image_len, &off, &len,
-		&ret_offset, kname);
+	    isfile = parse_executable (image_base, image_len, &off, &len,
+				       &ret_offset, kname);
 	}
     }
 
@@ -1187,24 +1253,62 @@ int bootmain (void)
     	params = params_device;
     	params_device = sol_params;
     } else if (!other) {
+	struct HdrS_struct *hdrs;
+
     	params_device = 0;
-        memcpy (image_base, image_base + off, len);
-        p = find_linux_HdrS (image_base, image_len);
-        if (p) {
-            if (fill_reboot_cmd && *(unsigned short *)(p + 8) >= 0x201) { /* ie. uses reboot_command */
-                char *q = (char *)(*(unsigned int *)(p + 24)), *r;
-                extern char bootdevice[];
+
+	memcpy (image_base, image_base + off, len);
+
+        hdrs = (struct HdrS_struct *)
+		silo_find_linux_HdrS((char *)image_base, image_len);
+
+	if (hdrs && hdrs->ver < 0x300 && image_base != (unsigned char *)0x4000) {
+	    /* Kernel doesn't support being loaded to other than
+	     * phys_base, so let's try to copy it down there. */
+	    if ((unsigned int)&_start - 0x4000 < len) {
+		/* Fuck, can't do that */
+		printf("Your kernel cannot fit into the memory destination. This\n"
+		       "can be resolved by recompiling the kernel with more devices\n"
+		       "built as modules, or upgrading your kernel to one that\n"
+		       "supports being loaded to higher memory areas (currently\n"
+		       "2.6.3+ or 2.4.26+).\n");
+		goto try_again;
+	    }
+
+	    printf("Kernel doesn't support loading to high memory, relocating...");
+
+	    /* Ok, it fits, so copy it down there */
+	    memcpy ((char *)0x4000, image_base, len);
+	    image_base = (unsigned char *)0x4000;
+
+	    /* Readjust some things */
+	    ret_offset = 0x4000;
+	    hdrs = (struct HdrS_struct *)
+		silo_find_linux_HdrS((char *)image_base, image_len);
+
+	    printf("done.\n");
+	}
+
+        if (hdrs) {
+	    unsigned int linux_ver = hdrs->linux_ver;
+
+	    printf("Loaded kernel version %d.%d.%d\n", (linux_ver >> 16) & 0xff,
+		   (linux_ver >> 8) & 0xff, linux_ver & 0xff);
+
+            if (fill_reboot_cmd && hdrs->ver >= 0x201) { /* ie. uses reboot_command */
+		char *q = (char *)hdrs->reboot_cmd_ptr_high;
+                char *r;
 
                 /* On Ultra there is xword there, this hack makes
                  * it work...
                  */
                 if (q == (char *)0xfffff800 || !q)
-                       q = (char *)(*(unsigned int *)(p+28));
+                       q = (char *)hdrs->reboot_cmd_ptr_low;
                 q = (char *)(((unsigned long)q)& 0x003fffff);
                 if (q >= (char *)0x4000 && q <= (char *)0x300000) {
                     if (given_bootargs_by_user) {
-                        if (strlen (bootdevice) <= 254) {
-			    strcpy (q, bootdevice);
+                        if (strlen (silo_disk_get_bootdevice()) <= 254) {
+			    strcpy (q, silo_disk_get_bootdevice());
 			    r = strchr (q, 0);
 			    if (strlen (given_bootargs) < 255 - (r - q)) {
 			        *r++ = ' ';
@@ -1215,7 +1319,7 @@ int bootmain (void)
                         strcpy (q, given_bootargs);
                 }
             }
-            if (!dig_into_params (params) && !*(unsigned short *)(p + 12)) {
+            if (!dig_into_params (params) && !hdrs->root_dev) {
                 char *s1, *s2;
 
                 s1 = cfg_get_strg(0, "root");
@@ -1227,12 +1331,15 @@ int bootmain (void)
                     params = s2;
                 }
             }
-            if (*(unsigned short *)(p + 8) >= 0x202)
-            	kernel_params = (char *)(*(unsigned int *)(p + 36)&0x3fffff);
-            if (*(unsigned short *)(p + 8) >= 0x203) {
-		linux_end = *(unsigned int *)(p + 40)&0x3fffff;
-		linux_end = (linux_end + 32 + 0x1fff) & ~0x1fff;
+
+            if (hdrs->ver >= 0x202) {
+		if (architecture == sun4u)
+		    kernel_params = (char *)((hdrs->bootstr_info_ptr_low - 0x400000) + 
+				(image_base - 0x4000));
+		else
+		    kernel_params = (char *)(hdrs->bootstr_info_ptr_low & 0x3fffff);
 	    }
+
             /* Some UltraAX machines have /dev/fd1 floppies only. */
             if (floppyswap) {
             	char *s1;
@@ -1240,22 +1347,30 @@ int bootmain (void)
             	for (s1 = params; (s1 = strstr(s1, "root=/dev/fd0")) != NULL; s1 += 13)
             		s1[12] = '1';
             }
+
             if (initrd_string) {
             	char *q, *r, *initrd_device, *initrd_kname, *initrd_limit, *initrd_cur, c;
 		char *string;
             	int initrd_partno, len, statusok = 0;
+
+		if (hdrs->ver >= 0x301)
+			initrd_can_do_64bit_phys = 1;
             	
             	q = strchr (initrd_string, '|');
             	if (q && !initrd_size) {
-            	    fatal ("When more than one initial ramdisk piece is specified, you have to give\n"
-            	           "a non-zero initrd-size option which is no shorter than sum of all pieces\n"
-            	           "lengths. Try again...\n");
+            	    silo_fatal("When more than one initial ramdisk piece "
+			       "is specified, you have to give\n"
+			       "a non-zero initrd-size option which is no "
+			       "shorter than sum of all pieces\n"
+			       "lengths. Try again...\n");
             	    prom_halt ();
             	}
             	if (q) {
     		    initrd_start = memory_find ((initrd_size + 16383) & ~16383);
     		    if (!initrd_start) {
-        		fatal ("You do not have enough continuous available memory for such initial ramdisk.");
+        		silo_fatal("You do not have enough continuous "
+				   "available memory for such initial "
+				   "ramdisk.");
         		prom_halt ();
     		    }
     		    string = strdup (initrd_string);
@@ -1270,55 +1385,76 @@ int bootmain (void)
 			parse_name (r, initrd_defpart, &initrd_device, &initrd_partno, &initrd_kname);
 			if (!initrd_kname) break;
 			if (!initrd_device) initrd_device = initrd_defdevice;
-			if (!load_file (initrd_device, initrd_partno, initrd_kname, initrd_cur, initrd_limit, &len, 0, 0)) break;
+			if (!silo_load_file(initrd_device, initrd_partno,
+					    initrd_kname, (unsigned char *)initrd_cur,
+					    (unsigned char *)initrd_limit, &len, 0, 0))
+			  break;
+			initrd_cur += len;
 			if (!c) {
 			    statusok = 1;
 			    break;
 			}
-			initrd_cur += len;
 			r = q + 1;
 			q = strchr (r, '|');
 			if (!q) q = strchr (r, 0);
 			if (initrd_prompt) {
-			    close ();
+			    silo_disk_close();
 			    printf ("Insert next media and press ENTER");
 			    prom_getchar ();
 			    printf ("\n");
 			}
     		    }
     		    free (string);
+		    printf("Loaded initial ramdisk (%d bytes at 0x%x)...\n", (unsigned int)initrd_cur -
+			   (unsigned int)initrd_start, initrd_start);
     		    if (statusok) {
-	        	extern unsigned long sun4u_initrd_pa;
+	        	extern unsigned long long sun4u_initrd_phys;
 	        	extern unsigned long sun4m_initrd_pa;
-	        	if ((linux_end & 0x3fffff) == (unsigned int)initrd_start)
-			    *(unsigned int *)(p + 16) = linux_end;
-			else if (architecture == sun4u)
-	            	    *(unsigned int *)(p + 16) = ((unsigned int)sun4u_initrd_pa + 0x400000);
-	            	else if (sun4m_initrd_pa)
-	            	    *(unsigned int *)(p + 16) = ((unsigned int)sun4m_initrd_pa);
-	            	else
-	            	    *(unsigned int *)(p + 16) = ((unsigned int)initrd_start | 0xf0000000);
-	            	*(unsigned int *)(p + 20) = initrd_size;
+
+			if (architecture == sun4u) {
+			    if (initrd_can_do_64bit_phys) {
+				hdrs->ramdisk_image64 = 
+					sun4u_initrd_phys + 0x400000ULL;
+			    } else {
+				hdrs->ramdisk_image =
+					(unsigned int)sun4u_initrd_phys + 0x400000;
+			    }
+			} else if (sun4m_initrd_pa) {
+	            	    hdrs->ramdisk_image = ((unsigned int)sun4m_initrd_pa);
+			} else
+	            	    hdrs->ramdisk_image = ((unsigned int)initrd_start | 0xf0000000);
+
+	            	hdrs->ramdisk_size = initrd_size;
     		    } else
     		        printf ("Error: initial ramdisk loading failed. No initrd will be used.\n");
             	} else {
 		    parse_name (initrd_string, initrd_defpart, &initrd_device, &initrd_partno, &initrd_kname);
 		    if (initrd_kname) {
 		        if (!initrd_device) initrd_device = initrd_defdevice;
-	        	if (load_file (initrd_device, initrd_partno, initrd_kname, (unsigned char *) 0x300000, (unsigned char *) 0x380000, 0, 0, initrd_lenfunc)) {
-	        	    extern unsigned long sun4u_initrd_pa;
+	        	if (silo_load_file(initrd_device, initrd_partno,
+					   initrd_kname, (unsigned char *) 0x300000,
+					   (unsigned char *) LARGE_RELOC,
+					   0, 0, initrd_lenfunc)) {
+	        	    extern unsigned long long sun4u_initrd_phys;
 	        	    extern unsigned long sun4m_initrd_pa;
-			    if ((linux_end & 0x3fffff) == (unsigned int)initrd_start)
-				*(unsigned int *)(p + 16) = linux_end;
-			    else if (architecture == sun4u)
-	            	        *(unsigned int *)(p + 16) = ((unsigned int)sun4u_initrd_pa + 0x400000);
-			    else if (sun4m_initrd_pa)
-				*(unsigned int *)(p + 16) = ((unsigned int)sun4m_initrd_pa);
-	            	    else
-	            	        *(unsigned int *)(p + 16) = ((unsigned int)initrd_start | 0xf0000000);
-	            	    *(unsigned int *)(p + 20) = initrd_size;
+
+			    if (architecture == sun4u) {
+				if (initrd_can_do_64bit_phys) {
+				    hdrs->ramdisk_image64 =
+					sun4u_initrd_phys + 0x400000ULL;
+				} else {
+				    hdrs->ramdisk_image =
+					(unsigned int)sun4u_initrd_phys + 0x400000;
+				}
+			    } else if (sun4m_initrd_pa) {
+				hdrs->ramdisk_image = ((unsigned int)sun4m_initrd_pa);
+			    } else
+	            	        hdrs->ramdisk_image = ((unsigned int)initrd_start | 0xf0000000);
+
+	            	    hdrs->ramdisk_size = initrd_size;
 	        	}
-		    }
+		    } else
+			printf ("Error: initial ramdisk loading failed. No initrd will be used.\n");
 		}
 	    }
         }
@@ -1331,17 +1467,19 @@ int bootmain (void)
         strcpy (other_device, p);
         params_device = other_device;
     }
-    close ();
+
+    silo_disk_close();
     if (timer_status >= 1)
         close_timer ();
     if (proll) {
-	set_prollargs (params, (unsigned int)image_base, len);
+	silo_set_prollargs(params, (unsigned int)image_base, len);
 	if (show_arguments) {
 	    printf ("Arguments: \"%s\"\n");
     	    pause_after = 1;
 	}
     } else {
-	set_bootargs (params, params_device);
+	silo_set_bootargs(params, params_device);
+
 	if (kernel_params) {
     	    extern char barg_out[];
     	    int len = *(unsigned int *)kernel_params;
@@ -1350,22 +1488,27 @@ int bootmain (void)
     	    kernel_params [8 + len - 1] = 0;
     	    *(unsigned int *)(kernel_params + 4) = 1;
 	}
+
 	if (show_arguments) {
-    	    show_bootargs ();
+    	    silo_show_bootargs();
     	    pause_after = 1;
 	}
     }
+
     if (pause_after) {
         printf ("%s", pause_message);
         prom_getchar ();
         printf ("\n");
     }
+
     memory_release();
+
     if (other && reboot) {
         strcpy (sol_params, params_device);
         strcat (sol_params, " ");
         strcat (sol_params, params);
     	prom_reboot(sol_params);
     }
+
     return ret_offset;
 }
